import prometheus_client

from ..monitoring.metrics import latency_metric_for_fast_operation

metric_state_report_rpcs = prometheus_client.Counter(
    "state_report_rpcs",
    "Number of Executor state report RPCs to Server",
)
metric_state_report_rpc_errors = prometheus_client.Counter(
    "state_report_rpc_errors",
    "Number of Executor state report RPC errors",
)
metric_state_report_rpc_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "state_report_rpc", "Executor state report rpc to Server"
    )
)



import asyncio
import time
from typing import Any

from tensorlake.function_executor.proto.function_executor_pb2 import (
    ExecutionPlanUpdate,
    FunctionArg,
    FunctionCall,
    ReduceOp,
)

from indexify.executor.blob_store.blob_store import BLOBStore
from indexify.proto.executor_api_pb2 import (
    AllocationFailureReason,
    AllocationOutcomeCode,
)

from .allocation_info import AllocationInfo
from .allocation_input import AllocationInput
from .allocation_output import AllocationOutput
from .events import AllocationFinalizationFinished
from .metrics.finalize_allocation import (
    metric_allocation_finalization_errors,
    metric_allocation_finalization_latency,
    metric_allocation_finalizations,
    metric_allocations_finalizing,
)


async def finalize_allocation(
    alloc_info: AllocationInfo, blob_store: BLOBStore, logger: Any
) -> AllocationFinalizationFinished:
    """Prepares the alloc output for getting it reported to Server.

    The alloc output is either coming from a failed alloc or from its finished execution on the Function Executor.
    Doesn't raise any Exceptions.
    """
    logger = logger.bind(module=__name__)
    start_time = time.monotonic()

    with (
        metric_allocations_finalizing.track_inprogress(),
        metric_allocation_finalization_latency.time(),
        metric_allocation_finalization_errors.count_exceptions(),
    ):
        metric_allocation_finalizations.inc()
        try:
            await _finalize_alloc_output(
                alloc_info=alloc_info,
                blob_store=blob_store,
                logger=logger,
            )
            logger.info(
                "allocation finalized",
                duration=time.monotonic() - start_time,
            )
            return AllocationFinalizationFinished(
                alloc_info=alloc_info, is_success=True
            )
        except asyncio.CancelledError:
            return AllocationFinalizationFinished(
                alloc_info=alloc_info, is_success=False
            )
        except BaseException as e:
            logger.error(
                "failed to finalize allocation",
                exc_info=e,
                duration=time.monotonic() - start_time,
            )
            return AllocationFinalizationFinished(
                alloc_info=alloc_info, is_success=False
            )


class _AllocationOutputSummary:
    def __init__(self):
        self.values_count: int = 0
        self.values_bytes: int = 0
        self.function_calls_count: int = 0
        self.reducer_calls_count: int = 0
        self.is_request_error: bool = False
        self.request_error_bytes: int = 0


async def _finalize_alloc_output(
    alloc_info: AllocationInfo, blob_store: BLOBStore, logger: Any
) -> None:
    """Finalizes the allocation output.

    Raises exception on error."""
    if alloc_info.input is None:
        raise Exception(
            "allocation input is None, this should never happen",
        )
    if alloc_info.output is None:
        raise Exception(
            "allocation output is None, this should never happen",
        )

    input: AllocationInput = alloc_info.input
    output: AllocationOutput = alloc_info.output

    output_summary: _AllocationOutputSummary = _alloc_output_summary(output)
    logger.info(
        "allocation output summary",
        values_count=output_summary.values_count,
        values_bytes=output_summary.values_bytes,
        function_calls_count=output_summary.function_calls_count,
        reducer_calls_count=output_summary.reducer_calls_count,
        is_request_error=output_summary.is_request_error,
        request_error_bytes=output_summary.request_error_bytes,
    )

    _log_function_metrics(output, logger)

    if output.outcome_code == AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_SUCCESS:
        if len(output.uploaded_function_outputs_blob.chunks) == 0:
            # A function always returns at least None so this branch should never be taken
            # unless FE misbehaved. No outputs were uploaded by function so
            # complete_multipart_upload will fail. So we abort the upload instead.
            await blob_store.abort_multipart_upload(
                uri=input.function_outputs_blob_uri,
                upload_id=input.function_outputs_blob_upload_id,
                logger=logger,
            )
        else:
            await blob_store.complete_multipart_upload(
                uri=input.function_outputs_blob_uri,
                upload_id=input.function_outputs_blob_upload_id,
                parts_etags=[
                    blob_chunk.etag
                    for blob_chunk in output.uploaded_function_outputs_blob.chunks
                ],
                logger=logger,
            )
        await blob_store.abort_multipart_upload(
            uri=input.request_error_blob_uri,
            upload_id=input.request_error_blob_upload_id,
            logger=logger,
        )
    elif output.outcome_code == AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE:
        await blob_store.abort_multipart_upload(
            uri=input.function_outputs_blob_uri,
            upload_id=input.function_outputs_blob_upload_id,
            logger=logger,
        )
        if (
            output.failure_reason
            == AllocationFailureReason.ALLOCATION_FAILURE_REASON_REQUEST_ERROR
        ):
            if len(output.uploaded_request_error_blob.chunks) == 0:
                # A function that fails with request error always returns at least one request error output
                # so this branch should never be taken unless FE misbehaved. No request error outputs were
                # uploaded by function so complete_multipart_upload will fail. So we abort the upload instead.
                await blob_store.abort_multipart_upload(
                    uri=input.request_error_blob_uri,
                    upload_id=input.request_error_blob_upload_id,
                    logger=logger,
                )
            else:
                await blob_store.complete_multipart_upload(
                    uri=input.request_error_blob_uri,
                    upload_id=input.request_error_blob_upload_id,
                    parts_etags=[
                        blob_chunk.etag
                        for blob_chunk in output.uploaded_request_error_blob.chunks
                    ],
                    logger=logger,
                )
        else:
            await blob_store.abort_multipart_upload(
                uri=input.request_error_blob_uri,
                upload_id=input.request_error_blob_upload_id,
                logger=logger,
            )
    else:
        raise ValueError(
            f"Unexpected alloction outcome code: {AllocationOutcomeCode.Name(output.outcome_code)}"
        )


def _alloc_output_summary(
    alloc_output: AllocationOutput,
) -> _AllocationOutputSummary:
    # self.values_count: int = 0
    # self.values_bytes: int = 0
    # self.function_calls_count: int = 0
    # self.reducer_calls_count: int = 0
    summary: _AllocationOutputSummary = _AllocationOutputSummary()

    if alloc_output.output_value is not None:
        summary.values_count += 1
        summary.values_bytes += alloc_output.output_value.manifest.size

    if alloc_output.output_execution_plan_updates is not None:
        for update in alloc_output.output_execution_plan_updates.updates:
            update: ExecutionPlanUpdate
            if update.HasField("function_call"):
                summary.function_calls_count += 1
                func_call: FunctionCall = update.function_call
                for arg in func_call.args:
                    arg: FunctionArg
                    if arg.HasField("value"):
                        summary.values_count += 1
                        summary.values_bytes += arg.value.manifest.size
            elif update.HasField("reduce"):
                summary.reducer_calls_count += 1
                reduce_op: ReduceOp = update.reduce
                for arg in reduce_op.collection:
                    arg: FunctionArg
                    if arg.HasField("value"):
                        summary.values_count += 1
                        summary.values_bytes += arg.value.manifest.size

    if alloc_output.request_error_output is not None:
        summary.is_request_error = True
        summary.request_error_bytes = alloc_output.request_error_output.manifest.size

    return summary


# Temporary workaround is logging customer metrics until we store them somewhere
# for future retrieval and processing.
def _log_function_metrics(output: AllocationOutput, logger: Any):
    if output.metrics is None:
        return

    for counter_name, counter_value in output.metrics.counters.items():
        logger.info(
            "function_metric", counter_name=counter_name, counter_value=counter_value
        )
    for timer_name, timer_value in output.metrics.timers.items():
        logger.info("function_metric", timer_name=timer_name, timer_value=timer_value)



from dataclasses import dataclass
from typing import Any, List

from .function_executor_server import FunctionExecutorServer


@dataclass
class FunctionExecutorServerConfiguration:
    """Configuration for creating a FunctionExecutorServer.

    This configuration only includes data that must be known
    during creation of the FunctionExecutorServer. If some data
    is not required during the creation then it shouldn't be here.

    A particular factory implementation might ignore certain
    configuration parameters or raise an exception if it can't implement
    them."""

    executor_id: str
    function_executor_id: str
    namespace: str
    application_name: str
    function_name: str
    application_version: str
    secret_names: List[str]
    cpu_ms_per_sec: int
    memory_bytes: int
    disk_bytes: int
    gpu_count: int


class FunctionExecutorServerFactory:
    """Abstract class for creating FunctionExecutorServers."""

    async def create(
        self, config: FunctionExecutorServerConfiguration, logger: Any
    ) -> FunctionExecutorServer:
        """Creates a new FunctionExecutorServer.

        Raises an exception if the creation failed or the configuration is not supported.
        Args:
            config: configuration of the FunctionExecutorServer.
            logger: logger to be used during the function call."""
        raise NotImplementedError()

    async def destroy(self, server: FunctionExecutorServer, logger: Any) -> None:
        """Destroys the FunctionExecutorServer and release all its resources.

        Args:
            logger: logger to be used during the function call.
        FunctionExecutorServer and customer code that it's running are not notified about the destruction.
        Never raises any Exceptions."""
        raise NotImplementedError



import asyncio
import os
import os.path
from typing import Any


class LocalFSBLOBStore:
    """BLOB store that stores BLOBs in local file system."""

    async def get(self, uri: str, logger: Any) -> bytes:
        """Returns binary value stored in file at the supplied URI.

        The URI must be a file URI (starts with "file://"). The path must be absolute.
        Raises Exception on error. Raises KeyError if the file doesn't exist.
        """
        # Run synchronous code in a thread to not block the event loop.
        return await asyncio.to_thread(self._sync_get, _path_from_file_uri(uri))

    async def presign_get_uri(self, uri: str, expires_in_sec: int, logger: Any) -> str:
        """Returns a presigned URI for getting the file at the supplied URI.

        For local files, just returns the file URI itself.
        """
        return uri

    async def upload(self, uri: str, value: bytes, logger: Any) -> None:
        """Stores the supplied binary value in a file at the supplied URI.

        The URI must be a file URI (starts with "file://"). The path must be absolute.
        Overwrites existing file. Raises Exception on error.
        """
        # Run synchronous code in a thread to not block the event loop.
        return await asyncio.to_thread(self._sync_put, _path_from_file_uri(uri), value)

    async def create_multipart_upload(self, uri: str, logger: Any) -> str:
        """Creates a multipart upload for local file and returns a dummy upload ID."""
        # Local files do not require multipart upload, return a dummy ID
        return "local-multipart-upload-id"

    async def complete_multipart_upload(
        self, uri: str, upload_id: str, parts_etags: list[str], logger: Any
    ) -> None:
        """Completes a multipart upload for local file. No-op for local files."""
        # No action needed for local files
        return None

    async def abort_multipart_upload(
        self, uri: str, upload_id: str, logger: Any
    ) -> None:
        """Aborts a multipart upload for local file. No-op for local files."""
        # No action needed for local files
        return None

    async def presign_upload_part_uri(
        self,
        uri: str,
        part_number: int,
        upload_id: str,
        expires_in_sec: int,
        logger: Any,
    ) -> str:
        """Returns a presigned URI for uploading a part in a multipart upload for local file.

        For local files, just returns the file URI itself.
        """
        return uri

    def _sync_get(self, path: str) -> bytes:
        if not os.path.isabs(path):
            raise ValueError(f"Path {path} must be absolute")

        if os.path.exists(path):
            with open(path, mode="rb") as blob_file:
                return blob_file.read()
        else:
            raise KeyError(f"File at {path} does not exist")

    def _sync_put(self, path: str, value: bytes) -> None:
        if not os.path.isabs(path):
            raise ValueError(f"Path {path} must be absolute")

        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, mode="wb") as blob_file:
            blob_file.write(value)


def _path_from_file_uri(uri: str) -> str:
    return uri[7:]  # strip "file://" prefix



import prometheus_client

from ...monitoring.metrics import (
    latency_metric_for_customer_controlled_operation,
    latency_metric_for_fast_operation,
    latency_metric_for_slow_operation,
)

# This file contains all metrics used by FunctionExecutor.

metric_function_executors_count = prometheus_client.Gauge(
    "function_executors_count", "Number of existing Function Executors"
)

# Metrics about whole FE creation workflow.
metric_creations: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_creates",
    "Number of Function Executor creations",
)
metric_create_latency: prometheus_client.Histogram = (
    latency_metric_for_customer_controlled_operation(
        "function_executor_create", "Function Executor creation (aka cold start)"
    )
)
metric_create_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_create_errors", "Number of Function Executor creation errors"
)

# Metrics about whole FE destroy workflow.
metric_destroys: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_destroys", "Number of Function Executor destructions"
)
metric_destroy_latency: prometheus_client.Histogram = latency_metric_for_slow_operation(
    "function_executor_destroy", "Function Executor destruction"
)
metric_destroy_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_destroy_errors",
    "Number of Function Executor destruction errors, results in a resource leak",
)

# FE server create and destruction metrics.
metric_create_server_latency: prometheus_client.Histogram = (
    latency_metric_for_slow_operation(
        "function_executor_create_server", "Function Executor server creation"
    )
)
metric_create_server_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_create_server_errors",
    "Number of Function Executor server creation errors",
)
metric_destroy_server_latency: prometheus_client.Histogram = (
    latency_metric_for_slow_operation(
        "function_executor_destroy_server", "Function Executor server destruction"
    )
)
metric_destroy_server_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_destroy_server_errors",
    "Number of Function Executor server destruction errors",
)

# FE channel creation and destruction metrics.
metric_establish_channel_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "function_executor_establish_channel", "Function Executor channel establishment"
    )
)
metric_establish_channel_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_establish_channel_errors",
    "Number of Function Executor channel establishment errors",
)
metric_destroy_channel_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "function_executor_destroy_channel", "Function Executor channel destruction"
    )
)
metric_destroy_channel_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_destroy_channel_errors",
    "Number of Function Executor channel destruction errors",
)

# FE get_info RPC metrics.
metric_get_info_rpc_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "function_executor_get_info_rpc", "Function Executor get_info RPC"
    )
)
metric_get_info_rpc_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_get_info_rpc_errors",
    "Number of Function Executor get_info RPC errors",
)
metric_function_executor_infos: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_infos",
    "Number of Function Executor creations with particular info",
    ["version", "sdk_version", "sdk_language", "sdk_language_version"],
)

# FE initialization RPC metrics.
metric_initialize_rpc_latency: prometheus_client.Histogram = (
    latency_metric_for_customer_controlled_operation(
        "function_executor_initialize_rpc", "Function Executor initialize RPC"
    )
)
metric_initialize_rpc_errors: prometheus_client.Counter = prometheus_client.Counter(
    "function_executor_initialize_rpc_errors",
    "Number of Function Executor initialize RPC errors",
)

# FE request state client creation and destruction metrics.
metric_create_request_state_client_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "function_executor_create_request_state_client",
        "Function Executor request state client creation",
    )
)
metric_create_request_state_client_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "function_executor_create_request_state_client_errors",
        "Number of Function Executor request state client creation errors",
    )
)
metric_destroy_request_state_client_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "function_executor_destroy_request_state_client",
        "Function Executor request state client destruction",
    )
)
metric_destroy_request_state_client_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "function_executor_destroy_request_state_client_errors",
        "Number of Function Executor request state client destruction errors",
    )
)

# FE health checker creation and destruction metrics.
metric_create_health_checker_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "function_executor_create_health_checker",
        "Function Executor health checker creation",
    )
)
metric_create_health_checker_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "function_executor_create_health_checker_errors",
        "Number of Function Executor health checker creation errors",
    )
)
metric_destroy_health_checker_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "function_executor_destroy_health_checker",
        "Function Executor health checker destruction",
    )
)
metric_destroy_health_checker_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "function_executor_destroy_health_checker_errors",
        "Number of Function Executor health checker destruction errors",
    )
)



import prometheus_client

from indexify.executor.monitoring.metrics import (
    latency_metric_for_customer_controlled_operation,
    latency_metric_for_fast_operation,
)

metric_control_loop_handle_event_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "handle_function_executor_control_loop_event",
        "Handle Function Executor control loop event",
    )
)

metric_allocations_fetched: prometheus_client.Counter = prometheus_client.Counter(
    "allocations_fetched",
    "Number of allocations that were fetched from Server",
)

metric_schedule_allocation_latency: prometheus_client.Histogram = (
    latency_metric_for_customer_controlled_operation(
        "schedule_allocation",
        "Schedule an allocation for execution after it got ready for execution",
    )
)
metric_runnable_allocations: prometheus_client.Gauge = prometheus_client.Gauge(
    "runnable_allocations",
    "Number of allocations that are ready for execution but are waiting to get scheduled to run on Function Executor (typically waiting for a free Function Executor)",
)
metric_runnable_allocations_per_function_name: prometheus_client.Gauge = (
    prometheus_client.Gauge(
        "runnable_allocations_per_function_name",
        "Number of allocations that are ready for execution but are waiting to get scheduled to run on Function Executor (typically waiting for a free Function Executor)",
        ["function_name"],
    )
)

metric_function_executors_with_state: prometheus_client.Gauge = prometheus_client.Gauge(
    "function_executors_with_state",
    "Number of Function Executors with a particular internal state",
    ["state"],
)
METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_UNKNOWN = "unknown"
METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_NOT_STARTED = "not_started"
METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_STARTING_UP = "starting_up"
METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_RUNNING = "running"
METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_TERMINATING = "terminating"
METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_TERMINATED = "terminated"


metric_function_executors_with_state.labels(
    state=METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_UNKNOWN
)
metric_function_executors_with_state.labels(
    state=METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_NOT_STARTED
)
metric_function_executors_with_state.labels(
    state=METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_STARTING_UP
)
metric_function_executors_with_state.labels(
    state=METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_RUNNING
)
metric_function_executors_with_state.labels(
    state=METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_TERMINATING
)
metric_function_executors_with_state.labels(
    state=METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_TERMINATED
)



from aiohttp import web

from .handler import Handler


class MonitoringServer:
    def __init__(
        self,
        host: str,
        port: int,
        startup_probe_handler: Handler,
        health_probe_handler: Handler,
        metrics_handler: Handler,
        reported_state_handler: Handler,
        desired_state_handler: Handler,
    ):
        self._host = host
        self._port = port
        self._app: web.Application = web.Application()
        self._app.add_routes(
            [
                web.get("/monitoring/startup", startup_probe_handler.handle),
                web.get("/monitoring/health", health_probe_handler.handle),
                web.get("/monitoring/metrics", metrics_handler.handle),
                web.get("/state/reported", reported_state_handler.handle),
                web.get("/state/desired", desired_state_handler.handle),
            ]
        )
        self._app_runner: web.AppRunner = web.AppRunner(self._app)

    async def run(self):
        await self._app_runner.setup()
        site = web.TCPSite(
            runner=self._app_runner,
            host=self._host,
            port=self._port,
            # Allow to listen when there's a closed socket in TIME_WAIT state
            reuse_address=True,
            # Don't allow other TCP sockets to actively listen on this address
            reuse_port=False,
        )
        await site.start()

    async def shutdown(self):
        await self._app_runner.cleanup()



import asyncio
import time
from typing import Any

import grpc
from tensorlake.function_executor.proto.function_executor_pb2 import (
    BLOB,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    Allocation as FEAllocation,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    AllocationFailureReason as FEAllocationFailureReason,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    AllocationOutcomeCode as FEAllocationOutcomeCode,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    AllocationResult,
    AwaitAllocationProgress,
    AwaitAllocationRequest,
    CreateAllocationRequest,
    DeleteAllocationRequest,
    SerializedObjectInsideBLOB,
)
from tensorlake.function_executor.proto.function_executor_pb2_grpc import (
    FunctionExecutorStub,
)
from tensorlake.function_executor.proto.message_validator import MessageValidator

from indexify.executor.function_executor.function_executor import FunctionExecutor
from indexify.executor.function_executor.health_checker import HealthCheckResult
from indexify.proto.executor_api_pb2 import (
    Allocation,
    AllocationFailureReason,
    AllocationOutcomeCode,
    FunctionExecutorTerminationReason,
)

from .allocation_info import AllocationInfo
from .allocation_output import AllocationMetrics, AllocationOutput
from .events import AllocationExecutionFinished
from .metrics.run_allocation import (
    metric_function_executor_run_allocation_rpc_errors,
    metric_function_executor_run_allocation_rpc_latency,
    metric_function_executor_run_allocation_rpcs,
    metric_function_executor_run_allocation_rpcs_in_progress,
)

_CREATE_ALLOCATION_TIMEOUT_SECS = 5
_DELETE_ALLOCATION_TIMEOUT_SECS = 5


async def run_allocation_on_function_executor(
    alloc_info: AllocationInfo, function_executor: FunctionExecutor, logger: Any
) -> AllocationExecutionFinished:
    """Runs the allocation on the Function Executor and sets alloc_info.output with the result.

    Doesn't raise any exceptions.
    """
    logger = logger.bind(module=__name__)

    if alloc_info.input is None:
        logger.error(
            "allocation input is None, this should never happen",
        )
        alloc_info.output = AllocationOutput.internal_error(
            allocation=alloc_info.allocation,
            execution_start_time=None,
            execution_end_time=None,
        )
        return AllocationExecutionFinished(
            alloc_info=alloc_info,
            function_executor_termination_reason=None,
        )

    fe_alloc: FEAllocation = FEAllocation(
        request_id=alloc_info.allocation.request_id,
        function_call_id=alloc_info.allocation.function_call_id,
        allocation_id=alloc_info.allocation.allocation_id,
        inputs=alloc_info.input.function_inputs,
    )

    function_executor.request_state_client().add_allocation_to_request_id_entry(
        allocation_id=alloc_info.allocation.allocation_id,
        request_id=alloc_info.allocation.request_id,
    )

    metric_function_executor_run_allocation_rpcs.inc()
    metric_function_executor_run_allocation_rpcs_in_progress.inc()
    # Not None if the Function Executor should be terminated after running the alloc.
    function_executor_termination_reason: FunctionExecutorTerminationReason | None = (
        None
    )

    # NB: We start this timer before invoking the first RPC, since
    # user code should be executing by the time the create_allocation() RPC
    # returns, so not attributing the allocation management RPC overhead to
    # the user would open a possibility for abuse. (This is somewhat
    # mitigated by the fact that these RPCs should have a very low
    # overhead.)
    execution_start_time: float = time.monotonic()

    # If this RPC failed due to customer code crashing the server we won't be
    # able to detect this. We'll treat this as our own error for now and thus
    # let the AioRpcError to be raised here.
    timeout_sec: float = alloc_info.allocation_timeout_ms / 1000.0
    try:
        # This aio task can only be cancelled during this await call.
        alloc_result: AllocationResult = await _run_allocation_rpcs(
            fe_alloc, function_executor, timeout_sec
        )
        alloc_info.output = _allocation_output_from_fe_result(
            allocation=alloc_info.allocation,
            result=alloc_result,
            execution_start_time=execution_start_time,
            execution_end_time=time.monotonic(),
            logger=logger,
        )
    except asyncio.TimeoutError:
        # This is an await_allocation() RPC timeout - we're not getting
        # progress messages or an allocation completion.
        function_executor_termination_reason = (
            FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_TIMEOUT
        )
        alloc_info.output = AllocationOutput.function_timeout(
            allocation=alloc_info.allocation,
            execution_start_time=execution_start_time,
            execution_end_time=time.monotonic(),
        )
    except grpc.aio.AioRpcError as e:
        # This indicates some sort of problem communicating with the FE.
        #
        # NB: We charge the user in these situations: code within the
        # FE is not isolated, so not charging would enable abuse.
        #
        # This is an unexpected situation, though, so we make sure to
        # log the situation for further investigation.

        function_executor_termination_reason = (
            FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_UNHEALTHY
        )
        metric_function_executor_run_allocation_rpc_errors.inc()

        if e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
            # This is either a create_allocation() RPC timeout or a
            # delete_allocation() RPC timeout; either suggests that the FE
            # is unhealthy.
            logger.error(
                "allocation management RPC execution deadline exceeded", exc_info=e
            )
        else:
            # This is a status from an unsuccessful RPC; this
            # shouldn't happen, but we handle it.
            logger.error("allocation management RPC failed", exc_info=e)

        alloc_info.output = AllocationOutput.function_executor_unresponsive(
            allocation=alloc_info.allocation,
            execution_start_time=execution_start_time,
            execution_end_time=time.monotonic(),
        )
    except asyncio.CancelledError:
        # Handle aio task cancellation during `await _run_allocation_rpcs`.
        # The allocation is still running in FE, we only cancelled the client-side RPC.
        function_executor_termination_reason = (
            FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_CANCELLED
        )
        alloc_info.output = AllocationOutput.allocation_cancelled(
            allocation=alloc_info.allocation,
            execution_start_time=execution_start_time,
            execution_end_time=time.monotonic(),
        )
    # TODO: Handle ValueError here separately. It happens when FE returned an invalid response
    # that doesn't pass validation. Still charge the user in this case to prevent service abuse.
    except Exception as e:
        # This is an unexpected exception; we believe that this
        # indicates an internal error.
        logger.error(
            "unexpected internal error during allocation lifecycle RPC sequence",
            exc_info=e,
        )
        alloc_info.output = AllocationOutput.internal_error(
            allocation=alloc_info.allocation,
            execution_start_time=execution_start_time,
            execution_end_time=time.monotonic(),
        )

    metric_function_executor_run_allocation_rpc_latency.observe(
        time.monotonic() - execution_start_time
    )
    metric_function_executor_run_allocation_rpcs_in_progress.dec()

    function_executor.request_state_client().remove_allocation_to_request_id_entry(
        allocation_id=alloc_info.allocation.allocation_id,
    )

    if (
        alloc_info.output.outcome_code
        == AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE
        and function_executor_termination_reason is None
    ):
        try:
            # Check if the allocation failed because the FE is unhealthy to prevent more allocations failing.
            result: HealthCheckResult = await function_executor.health_checker().check()
            if not result.is_healthy:
                function_executor_termination_reason = (
                    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_UNHEALTHY
                )
                logger.error(
                    "Function Executor health check failed after running allocation, shutting down Function Executor",
                    health_check_fail_reason=result.reason,
                )
        except asyncio.CancelledError:
            # The aio task was cancelled during the health check await.
            # We can't conclude anything about the health of the FE here.
            pass

    _log_alloc_execution_finished(output=alloc_info.output, logger=logger)

    return AllocationExecutionFinished(
        alloc_info=alloc_info,
        function_executor_termination_reason=function_executor_termination_reason,
    )


async def _run_allocation_rpcs(
    alloc: Allocation, function_executor: FunctionExecutor, timeout_sec: float
) -> AllocationResult:
    """Runs the allocation, returning the result, reporting errors via exceptions."""
    alloc_result: AllocationResult | None = None
    channel: grpc.aio.Channel = function_executor.channel()
    fe_stub = FunctionExecutorStub(channel)

    # Create allocation with timeout
    await fe_stub.create_allocation(
        CreateAllocationRequest(allocation=alloc),
        timeout=_CREATE_ALLOCATION_TIMEOUT_SECS,
    )

    # Await allocation with timeout resets on each response
    await_rpc = fe_stub.await_allocation(
        AwaitAllocationRequest(allocation_id=alloc.allocation_id)
    )

    try:
        while True:
            # Wait for next response with fresh timeout each time
            response: AwaitAllocationProgress = await asyncio.wait_for(
                await_rpc.read(), timeout=timeout_sec
            )

            if response == grpc.aio.EOF:
                break
            elif response.WhichOneof("response") == "allocation_result":
                alloc_result = response.allocation_result
                break

            # NB: We don't actually check for other message types
            # here; any message from the FE is treated as an
            # indication that it's making forward progress.
    finally:
        # Cancel the outstanding RPC to ensure any resources in use
        # are cleaned up; note that this is idempotent (in case the
        # RPC has already completed).
        await_rpc.cancel()

    # Delete allocation with timeout
    await fe_stub.delete_allocation(
        DeleteAllocationRequest(allocation_id=alloc.allocation_id),
        timeout=_DELETE_ALLOCATION_TIMEOUT_SECS,
    )

    if alloc_result is None:
        raise grpc.aio.AioRpcError(
            grpc.StatusCode.CANCELLED,
            None,
            None,
            "Function Executor didn't return allocation result",
        )

    return alloc_result


def _allocation_output_from_fe_result(
    allocation: Allocation,
    result: AllocationResult,
    execution_start_time: float,
    execution_end_time: float | None,
    logger: Any,
) -> AllocationOutput:
    response_validator = MessageValidator(result)
    response_validator.required_field("outcome_code")

    metrics = AllocationMetrics(counters={}, timers={})
    if result.HasField("metrics"):
        # Can be None if e.g. function failed.
        metrics.counters = dict(result.metrics.counters)
        metrics.timers = dict(result.metrics.timers)

    outcome_code: AllocationOutcomeCode = _to_alloc_outcome_code(
        result.outcome_code, logger=logger
    )
    failure_reason: AllocationFailureReason | None = None
    request_error_output: SerializedObjectInsideBLOB | None = None
    uploaded_request_error_blob: BLOB | None = None

    if outcome_code == AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE:
        response_validator.required_field("failure_reason")
        failure_reason: AllocationFailureReason | None = _to_alloc_failure_reason(
            result.failure_reason, logger
        )
        if (
            failure_reason
            == AllocationFailureReason.ALLOCATION_FAILURE_REASON_REQUEST_ERROR
        ):
            response_validator.required_field("request_error_output")
            response_validator.required_field("uploaded_request_error_blob")
            request_error_output = result.request_error_output
            uploaded_request_error_blob = result.uploaded_request_error_blob
    elif outcome_code == AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_SUCCESS:
        if not (result.HasField("value") or result.HasField("updates")):
            raise ValueError(
                "Function Executor returned AllocationResult with `success` outcome code but neither `value` nor `updates` are set in it"
            )
        response_validator.required_field("uploaded_function_outputs_blob")

    return AllocationOutput(
        allocation=allocation,
        outcome_code=outcome_code,
        failure_reason=failure_reason,
        output_value=result.value if result.HasField("value") else None,
        output_execution_plan_updates=(
            result.updates if result.HasField("updates") else None
        ),
        uploaded_function_outputs_blob=result.uploaded_function_outputs_blob,
        request_error_output=request_error_output,
        uploaded_request_error_blob=uploaded_request_error_blob,
        metrics=metrics,
        execution_start_time=execution_start_time,
        execution_end_time=execution_end_time,
    )


def _log_alloc_execution_finished(output: AllocationOutput, logger: Any) -> None:
    logger.info(
        "finished running allocation",
        success=output.outcome_code
        == AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_SUCCESS,
        outcome_code=AllocationOutcomeCode.Name(output.outcome_code),
        failure_reason=(
            AllocationFailureReason.Name(output.failure_reason)
            if output.failure_reason is not None
            else None
        ),
    )


def _to_alloc_outcome_code(
    fe_alloc_outcome_code: FEAllocationOutcomeCode, logger
) -> AllocationOutcomeCode:
    if fe_alloc_outcome_code == FEAllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_SUCCESS:
        return AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_SUCCESS
    elif (
        fe_alloc_outcome_code == FEAllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE
    ):
        return AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE
    else:
        logger.warning(
            "unknown AllocationOutcomeCode received from Function Executor",
            value=FEAllocationOutcomeCode.Name(fe_alloc_outcome_code),
        )
        return AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_UNKNOWN


def _to_alloc_failure_reason(
    fe_alloc_failure_reason: FEAllocationFailureReason, logger: Any
) -> AllocationFailureReason:
    if (
        fe_alloc_failure_reason
        == FEAllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_ERROR
    ):
        return AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_ERROR
    elif (
        fe_alloc_failure_reason
        == FEAllocationFailureReason.ALLOCATION_FAILURE_REASON_REQUEST_ERROR
    ):
        return AllocationFailureReason.ALLOCATION_FAILURE_REASON_REQUEST_ERROR
    elif (
        fe_alloc_failure_reason
        == FEAllocationFailureReason.ALLOCATION_FAILURE_REASON_INTERNAL_ERROR
    ):
        return AllocationFailureReason.ALLOCATION_FAILURE_REASON_INTERNAL_ERROR
    else:
        logger.warning(
            "unknown AllocationFailureReason received from Function Executor",
            value=FEAllocationFailureReason.Name(fe_alloc_failure_reason),
        )
        return AllocationFailureReason.ALLOCATION_FAILURE_REASON_UNKNOWN



import asyncio
from dataclasses import dataclass
from typing import Any

import grpc
from tensorlake.function_executor.proto.function_executor_pb2 import (
    InfoRequest,
    InfoResponse,
    InitializeRequest,
    InitializeResponse,
)
from tensorlake.function_executor.proto.function_executor_pb2_grpc import (
    FunctionExecutorStub,
)
from tensorlake.function_executor.proto.message_validator import MessageValidator
from tensorlake.utils.http_client import get_httpx_client

from indexify.executor.monitoring.metrics import IdempotentCounterChanger

from .health_checker import HealthChecker
from .metrics.function_executor import (
    metric_create_errors,
    metric_create_health_checker_errors,
    metric_create_health_checker_latency,
    metric_create_latency,
    metric_create_request_state_client_errors,
    metric_create_request_state_client_latency,
    metric_create_server_errors,
    metric_create_server_latency,
    metric_creations,
    metric_destroy_channel_errors,
    metric_destroy_channel_latency,
    metric_destroy_errors,
    metric_destroy_health_checker_errors,
    metric_destroy_health_checker_latency,
    metric_destroy_latency,
    metric_destroy_request_state_client_errors,
    metric_destroy_request_state_client_latency,
    metric_destroy_server_errors,
    metric_destroy_server_latency,
    metric_destroys,
    metric_establish_channel_errors,
    metric_establish_channel_latency,
    metric_function_executor_infos,
    metric_function_executors_count,
    metric_get_info_rpc_errors,
    metric_get_info_rpc_latency,
    metric_initialize_rpc_errors,
    metric_initialize_rpc_latency,
)
from .request_state_client import RequestStateClient
from .server.function_executor_server import (
    FUNCTION_EXECUTOR_SERVER_READY_TIMEOUT_SEC,
    FunctionExecutorServer,
)
from .server.function_executor_server_factory import (
    FunctionExecutorServerConfiguration,
    FunctionExecutorServerFactory,
)


@dataclass
class FunctionExecutorInitializationResult:
    """Result of FunctionExecutor initialization."""

    # If True, timed out waiting for the Function Executor to initialize.
    is_timeout: bool
    # FE is unresponsive if response is None.
    response: InitializeResponse | None


class FunctionExecutor:
    """Executor side class supporting a running FunctionExecutorServer.

    FunctionExecutor primary responsibility is creation and initialization
    of all resources associated with a particular Function Executor Server
    including the Server itself. FunctionExecutor owns all these resources
    and provides other Executor components with access to them.

    Addition of any business logic besides resource management is discouraged.
    Please add such logic to other classes managed by this class.
    """

    def __init__(self, server_factory: FunctionExecutorServerFactory, logger: Any):
        self._server_factory: FunctionExecutorServerFactory = server_factory
        self._logger: Any = logger.bind(module=__name__)
        self._server: FunctionExecutorServer | None = None
        self._channel: grpc.aio.Channel | None = None
        self._request_state_client: RequestStateClient | None = None
        self._health_checker: HealthChecker | None = None
        self._function_executors_counter_changer: IdempotentCounterChanger = (
            IdempotentCounterChanger(
                metric_function_executors_count,
            )
        )
        self._function_executors_counter_changer.inc()

    async def initialize(
        self,
        config: FunctionExecutorServerConfiguration,
        initialize_request: InitializeRequest,
        base_url: str,
        config_path: str | None,
        customer_code_timeout_sec: float,
    ) -> FunctionExecutorInitializationResult:
        """Creates and initializes a FunctionExecutorServer and all resources associated with it.

        Raises an Exception if an Executor side internal error occured."""
        try:
            with (
                metric_create_errors.count_exceptions(),
                metric_create_latency.time(),
            ):
                metric_creations.inc()
                await self._create_server(config)
                await self._establish_channel()
                stub: FunctionExecutorStub = FunctionExecutorStub(self._channel)
                await _collect_server_info(stub)
                await self._create_request_state_client(
                    stub=stub,
                    base_url=base_url,
                    config_path=config_path,
                    initialize_request=initialize_request,
                )
                await self._create_health_checker(self._channel, stub)

                return await _initialize_server(
                    stub, initialize_request, customer_code_timeout_sec, self._logger
                )
        except Exception:
            await self.destroy()
            raise

    def channel(self) -> grpc.aio.Channel:
        return self._channel

    def request_state_client(self) -> RequestStateClient:
        return self._request_state_client

    def health_checker(self) -> HealthChecker:
        return self._health_checker

    async def destroy(self):
        """Destroys all resources owned by this FunctionExecutor.

        Never raises any exceptions but logs them.
        Idempotent.
        """
        try:
            with (
                metric_destroy_errors.count_exceptions(),
                metric_destroy_latency.time(),
            ):
                self._function_executors_counter_changer.dec()
                metric_destroys.inc()
                await self._destroy_health_checker()
                await self._destroy_request_state_client()
                await self._destroy_channel()
                await self._destroy_server()
        except Exception as e:
            self._logger.error(
                "exception from a Function Executor destroy step, some destroy steps are not executed, this is a resource leak",
                exc_info=e,
            )

    async def _create_server(self, config: FunctionExecutorServerConfiguration) -> None:
        with (
            metric_create_server_errors.count_exceptions(),
            metric_create_server_latency.time(),
        ):
            self._server = await self._server_factory.create(
                config=config, logger=self._logger
            )

    async def _destroy_server(self) -> None:
        if self._server is None:
            return

        try:
            with (
                metric_destroy_server_errors.count_exceptions(),
                metric_destroy_server_latency.time(),
            ):
                await self._server_factory.destroy(self._server, self._logger)
        except Exception as e:
            self._logger.error("failed to destroy FunctionExecutorServer", exc_info=e)
        finally:
            self._server = None

    async def _establish_channel(self) -> None:
        with (
            metric_establish_channel_errors.count_exceptions(),
            metric_establish_channel_latency.time(),
        ):
            self._channel = await self._server.create_channel(self._logger)
            await asyncio.wait_for(
                self._channel.channel_ready(),
                timeout=FUNCTION_EXECUTOR_SERVER_READY_TIMEOUT_SEC,
            )

    async def _destroy_channel(self) -> None:
        if self._channel is None:
            return

        try:
            with (
                metric_destroy_channel_errors.count_exceptions(),
                metric_destroy_channel_latency.time(),
            ):
                await self._channel.close()
        except Exception as e:
            self._logger.error(
                "failed to close FunctionExecutorServer channel", exc_info=e
            )
        finally:
            self._channel = None

    async def _create_request_state_client(
        self,
        stub: FunctionExecutorStub,
        base_url: str,
        config_path: str | None,
        initialize_request: InitializeRequest,
    ) -> None:
        with (
            metric_create_request_state_client_errors.count_exceptions(),
            metric_create_request_state_client_latency.time(),
        ):
            self._request_state_client = RequestStateClient(
                stub=stub,
                base_url=base_url,
                http_client=get_httpx_client(config_path=config_path, make_async=True),
                application=initialize_request.function.application_name,
                namespace=initialize_request.function.namespace,
                logger=self._logger,
            )
            await self._request_state_client.start()

    async def _destroy_request_state_client(self) -> None:
        if self._request_state_client is None:
            return

        try:
            with (
                metric_destroy_request_state_client_errors.count_exceptions(),
                metric_destroy_request_state_client_latency.time(),
            ):
                await self._request_state_client.destroy()
        except Exception as e:
            self._logger.error(
                "failed to destroy FunctionExecutor invocation state client", exc_info=e
            )
        finally:
            self._request_state_client = None

    async def _create_health_checker(
        self, channel: grpc.aio.Channel, stub: FunctionExecutorStub
    ) -> None:
        with (
            metric_create_health_checker_errors.count_exceptions(),
            metric_create_health_checker_latency.time(),
        ):
            self._health_checker = HealthChecker(
                channel=channel,
                stub=stub,
                logger=self._logger,
            )

    async def _destroy_health_checker(self) -> None:
        if self._health_checker is None:
            return

        try:
            with (
                metric_destroy_health_checker_errors.count_exceptions(),
                metric_destroy_health_checker_latency.time(),
            ):
                self._health_checker.stop()
        except Exception as e:
            self._logger.error("failed to stop HealthChecker", exc_info=e)
        finally:
            self._health_checker = None


async def _collect_server_info(stub: FunctionExecutorStub) -> None:
    with (
        metric_get_info_rpc_errors.count_exceptions(),
        metric_get_info_rpc_latency.time(),
    ):
        info: InfoResponse = await stub.get_info(InfoRequest())
        validator = MessageValidator(info)
        validator.required_field("version")
        validator.required_field("sdk_version")
        validator.required_field("sdk_language")
        validator.required_field("sdk_language_version")

        metric_function_executor_infos.labels(
            version=info.version,
            sdk_version=info.sdk_version,
            sdk_language=info.sdk_language,
            sdk_language_version=info.sdk_language_version,
        ).inc()


async def _initialize_server(
    stub: FunctionExecutorStub,
    initialize_request: InitializeRequest,
    customer_code_timeout_sec: float,
    logger: Any,
) -> FunctionExecutorInitializationResult:
    with metric_initialize_rpc_latency.time():
        try:
            initialize_response: InitializeResponse = await stub.initialize(
                initialize_request,
                timeout=customer_code_timeout_sec,
            )
            return FunctionExecutorInitializationResult(
                is_timeout=False,
                response=initialize_response,
            )
        except grpc.aio.AioRpcError as e:
            # Increment the metric manually as we're not raising this exception.
            metric_initialize_rpc_errors.inc()
            metric_create_errors.inc()
            if e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                return FunctionExecutorInitializationResult(
                    is_timeout=True,
                    response=None,
                )
            else:
                logger.error(
                    "Function Executor initialize RPC failed",
                    exc_info=e,
                )
                return FunctionExecutorInitializationResult(
                    is_timeout=False,
                    response=None,
                )
